@@ -18,6 +18,7 @@
 #include "config.h"
 
 #include <math.h>
+#include <glib/gi18n.h>
 #include <gst/gst.h>
 #include <clapper-gtk/clapper-gtk.h>
 
@@ -44,6 +45,9 @@ struct _ClapperAppApplication
   gboolean need_init_state;
 };
 
+#define parent_class clapper_app_application_parent_class
+G_DEFINE_TYPE (ClapperAppApplication, clapper_app_application, GTK_TYPE_APPLICATION);
+
 struct ClapperPluginFeatureData
 {
   const gchar *name;
@@ -63,8 +67,147 @@ typedef struct
   const gchar *accels[3];
 } ClapperAppShortcut;
 
-#define parent_class clapper_app_application_parent_class
-G_DEFINE_TYPE (ClapperAppApplication, clapper_app_application, GTK_TYPE_APPLICATION);
+static gboolean
+clapper_app_options_get (const gchar *key, const gchar *format, GVariantDict *options,
+    GObject *src_object, GSettings *settings, gpointer output)
+{
+  if (options && g_variant_dict_lookup (options, key, format, output)) {
+    return TRUE;
+  } else if (src_object) {
+    g_object_get (src_object, key, output, NULL);
+    return TRUE;
+  } else if (settings) {
+    g_settings_get (settings, key, format, output);
+    return TRUE;
+  }
+
+  return FALSE;
+}
+
+static gboolean
+clapper_app_options_get_extra (const gchar *key, GVariantDict *options,
+    const gchar *extra_value, GSettings *settings, gchar **output)
+{
+  if (options && g_variant_dict_lookup (options, key, "s", output)) {
+    return TRUE;
+  } else if (extra_value) {
+    *output = g_strdup (extra_value);
+    return TRUE;
+  } else if (settings) {
+    *output = g_settings_get_string (settings, key);
+
+    /* Ensure non-empty string */
+    if (*output) {
+      if (strlen (*output) > 0)
+        return TRUE;
+      else
+        g_clear_pointer (output, g_free);
+    }
+  }
+
+  return FALSE;
+}
+
+/*
+ * Apply options to @dest_window. Option providers will be used in args order.
+ * If any arg is %NULL it will not be used. For example, passing %NULL as
+ * @settings will avoid restoring values to @dest_window from GSettings.
+ */
+static void
+clapper_app_apply_options_to_window (ClapperAppWindow *dest_window, GVariantDict *options,
+    ClapperAppWindow *src_window, GSettings *settings)
+{
+  ClapperPlayer *dest_player;
+  ClapperAppWindowExtraOptions *src_extra_opts = NULL, *dest_extra_opts = NULL;
+  GObject *src_player_obj = NULL;
+  GObject *src_queue_obj = NULL;
+  gchar *option_str;
+  gdouble option_dbl;
+  gint option_int;
+  gboolean option_bool;
+
+  GST_DEBUG ("Applying options to window: %p", dest_window);
+
+  dest_player = clapper_app_window_get_player (dest_window);
+  dest_extra_opts = clapper_app_window_get_extra_options (dest_window);
+
+  if (src_window) {
+    src_player_obj = (GObject *) clapper_app_window_get_player (src_window);
+    src_queue_obj = (GObject *) clapper_player_get_queue (CLAPPER_PLAYER_CAST (src_player_obj));
+    src_extra_opts = clapper_app_window_get_extra_options (src_window);
+  }
+
+  /* Apply player values, clamp them to be within allowed range */
+  if (clapper_app_options_get ("volume", "d", options, src_player_obj, settings, &option_dbl))
+    clapper_player_set_volume (dest_player, PERCENTAGE_ROUND (CLAMP (option_dbl, 0, 2.0)));
+  if (clapper_app_options_get ("mute", "b", NULL, src_player_obj, settings, &option_bool))
+    clapper_player_set_mute (dest_player, option_bool);
+  if (clapper_app_options_get ("speed", "d", options, src_player_obj, settings, &option_dbl))
+    clapper_player_set_speed (dest_player, PERCENTAGE_ROUND (CLAMP (option_dbl, 0.05, 2.0)));
+  if (clapper_app_options_get ("adaptive-start-bitrate", "i", options, src_player_obj, settings, &option_int))
+    clapper_player_set_adaptive_start_bitrate (dest_player, option_int);
+  if (clapper_app_options_get ("progression-mode", "i", options, src_queue_obj, settings, &option_int))
+    clapper_queue_set_progression_mode (clapper_player_get_queue (dest_player), CLAMP (option_int, 0, 4));
+  if (clapper_app_options_get ("subtitles-enabled", "b", NULL, src_player_obj, settings, &option_bool))
+    clapper_player_set_subtitles_enabled (dest_player, option_bool);
+
+  if (clapper_app_options_get_extra ("video-filter", options,
+      (src_extra_opts) ? src_extra_opts->video_filter : NULL, NULL, &option_str)) {
+    clapper_player_set_video_filter (dest_player, clapper_app_utils_make_element (option_str));
+    g_free (dest_extra_opts->video_filter);
+    dest_extra_opts->video_filter = option_str;
+  }
+  if (clapper_app_options_get_extra ("audio-filter", options,
+      (src_extra_opts) ? src_extra_opts->audio_filter : NULL, NULL, &option_str)) {
+    clapper_player_set_audio_filter (dest_player, clapper_app_utils_make_element (option_str));
+    g_free (dest_extra_opts->audio_filter);
+    dest_extra_opts->audio_filter = option_str;
+  }
+  if (clapper_app_options_get_extra ("video-sink", options,
+      (src_extra_opts) ? src_extra_opts->video_sink : NULL, NULL, &option_str)) {
+    clapper_player_set_video_sink (dest_player, clapper_app_utils_make_element (option_str));
+    g_free (dest_extra_opts->video_sink);
+    dest_extra_opts->video_sink = option_str;
+  }
+  if (clapper_app_options_get_extra ("audio-sink", options,
+      (src_extra_opts) ? src_extra_opts->audio_sink : NULL, NULL, &option_str)) {
+    clapper_player_set_audio_sink (dest_player, clapper_app_utils_make_element (option_str));
+    g_free (dest_extra_opts->audio_sink);
+    dest_extra_opts->audio_sink = option_str;
+  }
+
+  /* Apply window options */
+  if ((options && g_variant_dict_contains (options, "fullscreen"))
+      || (settings && g_settings_get_boolean (settings, "fullscreened")))
+    gtk_window_fullscreen (GTK_WINDOW (dest_window));
+  else if (settings && g_settings_get_boolean (settings, "maximized"))
+    gtk_window_maximize (GTK_WINDOW (dest_window));
+
+  GST_DEBUG ("Options applied");
+}
+
+static inline void
+_store_settings_from_window (ClapperAppApplication *self, ClapperAppWindow *app_window)
+{
+  ClapperPlayer *player = clapper_app_window_get_player (app_window);
+  ClapperQueue *queue = clapper_player_get_queue (player);
+  GtkWindow *window = GTK_WINDOW (app_window);
+
+  GST_DEBUG ("Storing current configuration to GSettings");
+
+  g_settings_set_double (self->settings, "volume", clapper_player_get_volume (player));
+  g_settings_set_boolean (self->settings, "mute", clapper_player_get_mute (player));
+  g_settings_set_double (self->settings, "speed", clapper_player_get_speed (player));
+  g_settings_set_int (self->settings, "adaptive-start-bitrate",
+      CLAMP (clapper_player_get_adaptive_bandwidth (player) * 0.8, 0, G_MAXINT));
+  g_settings_set_boolean (self->settings, "subtitles-enabled", clapper_player_get_subtitles_enabled (player));
+  g_settings_set_int (self->settings, "progression-mode", clapper_queue_get_progression_mode (queue));
+
+  g_settings_set_boolean (self->settings, "maximized", gtk_window_is_maximized (window));
+  g_settings_set_boolean (self->settings, "fullscreened", gtk_window_is_fullscreen (window));
+
+  GST_DEBUG ("Configuration stored");
+}
 
 static inline void
 _set_initial_plugin_feature_ranks (void)
@@ -190,6 +333,43 @@ add_uri (GSimpleAction *action, GVariant *param, gpointer user_data)
   clapper_app_uri_dialog_open_uri (gtk_app);
 }
 
+static void
+clear_queue (GSimpleAction *action, GVariant *param, gpointer user_data)
+{
+  GtkApplication *gtk_app = GTK_APPLICATION (user_data);
+  GtkWindow *window = gtk_application_get_active_window (gtk_app);
+  ClapperPlayer *player;
+  ClapperQueue *queue;
+
+  while (window && !CLAPPER_APP_IS_WINDOW (window))
+    window = gtk_window_get_transient_for (window);
+
+  player = clapper_app_window_get_player (CLAPPER_APP_WINDOW (window));
+  queue = clapper_player_get_queue (player);
+
+  clapper_queue_clear (queue);
+}
+
+static void
+new_window (GSimpleAction *action, GVariant *param, gpointer user_data)
+{
+  GtkApplication *gtk_app = GTK_APPLICATION (user_data);
+  GtkWidget *stack = gtk_window_get_child (gtk_application_get_active_window (gtk_app));
+  const gchar *child_name = gtk_stack_get_visible_child_name (GTK_STACK (stack));
+
+  /* Do not allow to open new windows during initial state,
+   * there already is a free one to use */
+  if (g_strcmp0 (child_name, "initial_state") != 0) {
+    ClapperAppWindow *src_window, *dest_window;
+
+    src_window = CLAPPER_APP_WINDOW_CAST (gtk_application_get_active_window (gtk_app));
+    dest_window = CLAPPER_APP_WINDOW_CAST (clapper_app_window_new (gtk_app));
+
+    clapper_app_apply_options_to_window (dest_window, NULL, src_window, NULL);
+    gtk_window_present (GTK_WINDOW (dest_window));
+  }
+}
+
 static void
 show_preferences (GSimpleAction *action, GVariant *param, gpointer user_data)
 {
@@ -225,55 +405,12 @@ show_about (GSimpleAction *action, GVariant *param, gpointer user_data)
   gtk_window_present (GTK_WINDOW (about_window));
 }
 
-static inline void
-_restore_settings_to_window (ClapperAppApplication *self, ClapperAppWindow *app_window)
-{
-  ClapperPlayer *player = clapper_app_window_get_player (app_window);
-  ClapperQueue *queue = clapper_player_get_queue (player);
-
-  GST_DEBUG ("Restoring saved GSettings values to: %" GST_PTR_FORMAT, app_window);
-
-  clapper_player_set_volume (player, PERCENTAGE_ROUND (g_settings_get_double (self->settings, "volume")));
-  clapper_player_set_mute (player, g_settings_get_boolean (self->settings, "mute"));
-  clapper_player_set_speed (player, PERCENTAGE_ROUND (g_settings_get_double (self->settings, "speed")));
-  clapper_player_set_subtitles_enabled (player, g_settings_get_boolean (self->settings, "subtitles-enabled"));
-  clapper_queue_set_progression_mode (queue, g_settings_get_int (self->settings, "progression-mode"));
-
-  if (g_settings_get_boolean (self->settings, "fullscreened"))
-    gtk_window_fullscreen (GTK_WINDOW (app_window));
-  else if (g_settings_get_boolean (self->settings, "maximized"))
-    gtk_window_maximize (GTK_WINDOW (app_window));
-
-  GST_DEBUG ("Configuration restored");
-}
-
-static inline void
-_store_settings_from_window (ClapperAppApplication *self, ClapperAppWindow *app_window)
-{
-  ClapperPlayer *player = clapper_app_window_get_player (app_window);
-  ClapperQueue *queue = clapper_player_get_queue (player);
-  GtkWindow *window = GTK_WINDOW (app_window);
-
-  GST_DEBUG ("Storing current configuration to GSettings");
-
-  g_settings_set_double (self->settings, "volume", clapper_player_get_volume (player));
-  g_settings_set_boolean (self->settings, "mute", clapper_player_get_mute (player));
-  g_settings_set_double (self->settings, "speed", clapper_player_get_speed (player));
-  g_settings_set_boolean (self->settings, "subtitles-enabled", clapper_player_get_subtitles_enabled (player));
-  g_settings_set_int (self->settings, "progression-mode", clapper_queue_get_progression_mode (queue));
-
-  g_settings_set_boolean (self->settings, "maximized", gtk_window_is_maximized (window));
-  g_settings_set_boolean (self->settings, "fullscreened", gtk_window_is_fullscreen (window));
-
-  GST_DEBUG ("Configuration stored");
-}
-
 GApplication *
 clapper_app_application_new (void)
 {
   return g_object_new (CLAPPER_APP_TYPE_APPLICATION,
       "application-id", CLAPPER_APP_ID,
-      "flags", G_APPLICATION_HANDLES_OPEN,
+      "flags", G_APPLICATION_HANDLES_OPEN | G_APPLICATION_HANDLES_COMMAND_LINE,
       NULL);
 }
 
@@ -306,14 +443,18 @@ static void
 clapper_app_application_activate (GApplication *app)
 {
   ClapperAppApplication *self = CLAPPER_APP_APPLICATION_CAST (app);
+  GtkApplication *gtk_app = GTK_APPLICATION (app);
   GtkWindow *window;
 
   GST_INFO ("Activate");
   G_APPLICATION_CLASS (parent_class)->activate (app);
 
-  if (!(window = gtk_application_get_active_window (GTK_APPLICATION (app)))) {
-    window = GTK_WINDOW (clapper_app_window_new (GTK_APPLICATION (app)));
-    _restore_settings_to_window (self, CLAPPER_APP_WINDOW_CAST (window));
+  /* When activated through DBus command line does not run,
+   * so create our first window here instead */
+  if (!(window = gtk_application_get_active_window (gtk_app))) {
+    window = GTK_WINDOW (clapper_app_window_new (gtk_app));
+    clapper_app_apply_options_to_window (CLAPPER_APP_WINDOW_CAST (window),
+        NULL, NULL, self->settings);
   }
 
   if (self->need_init_state) {
@@ -331,6 +472,8 @@ clapper_app_application_local_command_line (GApplication *app,
   gchar **argv = *arguments;
   guint i;
 
+  GST_INFO ("Handling local command line");
+
   /* NOTE: argv is never NULL, so no need to check */
 
   for (i = 0; argv[i]; ++i) {
@@ -344,6 +487,43 @@ clapper_app_application_local_command_line (GApplication *app,
   return G_APPLICATION_CLASS (parent_class)->local_command_line (app, arguments, exit_status);
 }
 
+static gint
+clapper_app_application_command_line (GApplication *app, GApplicationCommandLine *cmd_line)
+{
+  ClapperAppApplication *self = CLAPPER_APP_APPLICATION_CAST (app);
+  ClapperAppWindow *src_window = NULL, *dest_window = NULL;
+  GtkApplication *gtk_app = GTK_APPLICATION (app);
+  GVariantDict *options;
+  GFile **files = NULL;
+  GSettings *settings;
+  gint n_files = 0;
+
+  GST_INFO ("Handling command line");
+
+  options = g_application_command_line_get_options_dict (cmd_line);
+  dest_window = CLAPPER_APP_WINDOW_CAST (gtk_application_get_active_window (gtk_app));
+
+  /* Restore settings only once by making them %NULL when run again */
+  settings = (!dest_window) ? self->settings : NULL;
+
+  if (!dest_window || g_variant_dict_contains (options, "new-window")) {
+    src_window = dest_window;
+    dest_window = CLAPPER_APP_WINDOW_CAST (clapper_app_window_new (gtk_app));
+  }
+
+  clapper_app_apply_options_to_window (dest_window, options, src_window, settings);
+
+  if (clapper_app_utils_files_from_command_line (cmd_line, &files, &n_files)) {
+    g_application_open (app, files, n_files,
+        (g_variant_dict_contains (options, "enqueue")) ? "add-only" : "");
+    clapper_app_utils_files_free (files);
+  } else {
+    g_application_activate (app);
+  }
+
+  return EXIT_SUCCESS;
+}
+
 static gboolean
 _is_claps_file (GFile *file)
 {
@@ -512,9 +692,25 @@ clapper_app_application_constructed (GObject *object)
   GApplication *app = G_APPLICATION (self);
   guint i;
 
-  static const GActionEntry app_entries[] = {
+  const GOptionEntry app_options[] = {
+    { "new-window", 'n', 0, G_OPTION_ARG_NONE, NULL, _("Create a new window"), NULL },
+    { "enqueue", 0, 0, G_OPTION_ARG_NONE, NULL, _("Add media to queue in primary application instance"), NULL },
+    { "volume", 0, 0, G_OPTION_ARG_DOUBLE, NULL, _("Audio volume to set (0 - 2.0 range)"), NULL },
+    { "speed", 0, 0, G_OPTION_ARG_DOUBLE, NULL, _("Playback speed to set (0.05 - 2.0 range)"), NULL },
+    { "adaptive-start-bitrate", 0, 0, G_OPTION_ARG_INT, NULL, _("Initial bitrate for adaptive streaming"), NULL },
+    { "progression-mode", 0, 0, G_OPTION_ARG_INT, NULL, _("Initial queue progression mode (0=none, 1=consecutive, 2=repeat-item, 3=carousel, 4=shuffle)"), NULL },
+    { "fullscreen", 'f', 0, G_OPTION_ARG_NONE, NULL, _("Set window to be fullscreen"), NULL },
+    { "video-filter", 0, 0, G_OPTION_ARG_STRING, NULL, _("Video filter to use (\"none\" to disable)"), NULL },
+    { "audio-filter", 0, 0, G_OPTION_ARG_STRING, NULL, _("Audio filter to use (\"none\" to disable)"), NULL },
+    { "video-sink", 0, 0, G_OPTION_ARG_STRING, NULL, _("Video sink to use"), NULL },
+    { "audio-sink", 0, 0, G_OPTION_ARG_STRING, NULL, _("Audio sink to use"), NULL },
+    { NULL }
+  };
+  static const GActionEntry app_actions[] = {
     { "add-files", add_files, NULL, NULL, NULL },
     { "add-uri", add_uri, NULL, NULL, NULL },
+    { "clear-queue", clear_queue, NULL, NULL, NULL },
+    { "new-window", new_window, NULL, NULL, NULL },
     { "info", show_info, NULL, NULL, NULL },
     { "preferences", show_preferences, NULL, NULL, NULL },
     { "about", show_about, NULL, NULL, NULL },
@@ -522,10 +718,13 @@ clapper_app_application_constructed (GObject *object)
   static const ClapperAppShortcut app_shortcuts[] = {
     { "app.add-files", { "<Control>o", NULL, NULL }},
     { "app.add-uri", { "<Control>u", NULL, NULL }},
+    { "app.new-window", { "<Control>n", NULL, NULL }},
     { "app.info", { "<Control>i", NULL, NULL }},
     { "app.preferences", { "<Control>comma", NULL, NULL }},
     { "app.about", { "F1", NULL, NULL }},
     { "win.toggle-fullscreen", { "F11", "f", NULL }},
+    { "win.unfullscreen", { "Escape", NULL, NULL }},
+    { "win.auto-resize", { "<Super>r", NULL, NULL }},
     { "win.show-help-overlay", { "<Control>question", NULL, NULL }},
     { "window.close", { "<Control>q", "q", NULL }},
   };
@@ -542,11 +741,14 @@ clapper_app_application_constructed (GObject *object)
   plugin_feature_ranks_settings_changed_cb (self->settings, NULL, NULL);
 
   g_action_map_add_action_entries (G_ACTION_MAP (app),
-      app_entries, G_N_ELEMENTS (app_entries), app);
+      app_actions, G_N_ELEMENTS (app_actions), app);
 
   for (i = 0; i < G_N_ELEMENTS (app_shortcuts); ++i)
     gtk_application_set_accels_for_action (GTK_APPLICATION (app), app_shortcuts[i].action, app_shortcuts[i].accels);
 
+  g_application_set_option_context_parameter_string (app, "[URI1|FILE1] [URI2|FILE2] …");
+
+  g_application_add_main_option_entries (app, app_options);
   g_application_add_option_group (app, gst_init_get_option_group ());
 
   G_OBJECT_CLASS (parent_class)->constructed (object);
@@ -581,5 +783,6 @@ clapper_app_application_class_init (ClapperAppApplicationClass *klass)
 
   application_class->activate = clapper_app_application_activate;
   application_class->local_command_line = clapper_app_application_local_command_line;
+  application_class->command_line = clapper_app_application_command_line;
   application_class->open = clapper_app_application_open;
 }
