@@ -28,6 +28,9 @@
 #include "clapper-app-file-dialog.h"
 #include "clapper-app-utils.h"
 
+#define MIN_WINDOW_WIDTH 352
+#define MIN_WINDOW_HEIGHT 198
+
 #define DEFAULT_WINDOW_WIDTH 1024
 #define DEFAULT_WINDOW_HEIGHT 576
 
@@ -40,6 +43,8 @@
 #define PERCENTAGE_ROUND(a) (round ((gdouble) a / 0.01) * 0.01)
 #define AXIS_WINS_OVER(a,b) ((a > 0 && a - 0.3 > b) || (a < 0 && a + 0.3 < b))
 
+#define MIN_STEP_DELAY 12000
+
 #define GST_CAT_DEFAULT clapper_app_window_debug
 GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);
 
@@ -59,6 +64,7 @@ struct _ClapperAppWindow
   GSettings *settings;
 
   guint seek_timeout;
+  guint resize_tick_id;
 
   gboolean key_held;
   gboolean scrolling;
@@ -74,6 +80,36 @@ struct _ClapperAppWindow
 #define parent_class clapper_app_window_parent_class
 G_DEFINE_TYPE (ClapperAppWindow, clapper_app_window, GTK_TYPE_APPLICATION_WINDOW)
 
+typedef struct
+{
+  gint dest_width, dest_height;
+  gint64 last_tick;
+} ClapperAppWindowResizeData;
+
+#if CLAPPER_HAVE_MPRIS
+static guint16 instance_count = 0;
+#endif
+
+static inline GQuark
+clapper_app_window_extra_options_get_quark (void)
+{
+  return g_quark_from_static_string ("clapper-app-window-extra-options-quark");
+}
+
+static void
+clapper_app_window_extra_options_free (ClapperAppWindowExtraOptions *extra_opts)
+{
+  GST_TRACE ("Freeing window extra options: %p", extra_opts);
+
+  g_free (extra_opts->video_filter);
+  g_free (extra_opts->audio_filter);
+
+  g_free (extra_opts->video_sink);
+  g_free (extra_opts->audio_sink);
+
+  g_free (extra_opts);
+}
+
 static void
 _media_item_title_changed_cb (ClapperMediaItem *item,
     GParamSpec *pspec G_GNUC_UNUSED, ClapperAppWindow *self)
@@ -115,6 +151,15 @@ _queue_current_item_changed_cb (ClapperQueue *queue,
   gst_clear_object (&current_item);
 }
 
+static void
+_player_adaptive_bandwidth_changed_cb (ClapperPlayer *player,
+    GParamSpec *pspec G_GNUC_UNUSED, gpointer *user_data G_GNUC_UNUSED)
+{
+  /* Do not take whole bandwidth */
+  clapper_player_set_adaptive_start_bitrate (player,
+      clapper_player_get_adaptive_bandwidth (player) * 0.8);
+}
+
 static gboolean
 _get_seek_method_mapping (GValue *value,
     GVariant *variant, gpointer user_data G_GNUC_UNUSED)
@@ -224,12 +269,16 @@ _open_subtitles_cb (ClapperGtkExtraMenuButton *button G_GNUC_UNUSED,
 }
 
 static void
-right_click_pressed_cb (GtkGestureClick *click, gint n_press,
+click_pressed_cb (GtkGestureClick *click, gint n_press,
     gdouble x, gdouble y, ClapperAppWindow *self)
 {
   GdkCursor *cursor;
   const gchar *cursor_name = NULL;
 
+  if (gtk_gesture_single_get_current_button (
+      GTK_GESTURE_SINGLE (click)) != GDK_BUTTON_SECONDARY)
+    return;
+
   GST_LOG_OBJECT (self, "Right click pressed");
 
   if ((cursor = gtk_widget_get_cursor (self->video)))
@@ -244,9 +293,152 @@ right_click_pressed_cb (GtkGestureClick *click, gint n_press,
   }
 }
 
+static gboolean
+_resize_tick (GtkWidget *widget, GdkFrameClock *frame_clock,
+    ClapperAppWindowResizeData *resize_data)
+{
+  gint64 now = gdk_frame_clock_get_frame_time (frame_clock);
+
+  if (now - resize_data->last_tick >= MIN_STEP_DELAY) {
+    ClapperAppWindow *self = CLAPPER_APP_WINDOW_CAST (widget);
+    gint win_width, win_height;
+
+    GST_LOG_OBJECT (self, "Resize step, last: %" G_GINT64_FORMAT
+        ", now: %" G_GINT64_FORMAT, resize_data->last_tick, now);
+
+    gtk_window_get_default_size (GTK_WINDOW (self), &win_width, &win_height);
+
+    if (win_width != resize_data->dest_width) {
+      gint width_diff = ABS (win_width - resize_data->dest_width);
+      gint step_size = (width_diff > 180) ? 120 : MAX (width_diff / 4, 1);
+
+      win_width += (win_width > resize_data->dest_width) ? -step_size : step_size;
+    }
+    if (win_height != resize_data->dest_height) {
+      gint height_diff = ABS (win_height - resize_data->dest_height);
+      gint step_size = (height_diff > 180) ? 120 : MAX (height_diff / 4, 1);
+
+      win_height += (win_height > resize_data->dest_height) ? -step_size : step_size;
+    }
+
+    gtk_window_set_default_size (GTK_WINDOW (self), win_width, win_height);
+
+    if (win_width == resize_data->dest_width
+        && win_height == resize_data->dest_height) {
+      GST_DEBUG_OBJECT (self, "Window resize finish");
+      self->resize_tick_id = 0;
+
+      return G_SOURCE_REMOVE;
+    }
+
+    resize_data->last_tick = now;
+  }
+
+  return G_SOURCE_CONTINUE;
+}
+
 static void
-right_click_released_cb (GtkGestureClick *click, gint n_press,
-    gdouble x, gdouble y, ClapperAppWindow *self)
+_calculate_win_resize (gint win_w, gint win_h,
+    gint vid_w, gint vid_h, gint *dest_w, gint *dest_h)
+{
+  gdouble win_aspect = (gdouble) win_w / win_h;
+  gdouble vid_aspect = (gdouble) vid_w / vid_h;
+
+  if (win_aspect < vid_aspect) {
+    while (!G_APPROX_VALUE (fmod (win_w, vid_aspect), 0, FLT_EPSILON))
+      win_w++;
+
+    win_h = round ((gdouble) win_w / vid_aspect);
+
+    if (win_h < MIN_WINDOW_HEIGHT) {
+      _calculate_win_resize (G_MAXINT, MIN_WINDOW_HEIGHT, vid_w, vid_h, dest_w, dest_h);
+      return;
+    }
+  } else {
+    while (!G_APPROX_VALUE (fmod (win_h * vid_aspect, 1.0), 0, FLT_EPSILON))
+      win_h++;
+
+    win_w = round ((gdouble) win_h * vid_aspect);
+
+    if (win_w < MIN_WINDOW_WIDTH) {
+      _calculate_win_resize (MIN_WINDOW_WIDTH, G_MAXINT, vid_w, vid_h, dest_w, dest_h);
+      return;
+    }
+  }
+
+  *dest_w = win_w;
+  *dest_h = win_h;
+}
+
+static void
+_resize_window (ClapperAppWindow *self)
+{
+  ClapperPlayer *player;
+  ClapperStreamList *vstreams;
+  ClapperVideoStream *vstream;
+  GdkToplevelState toplevel_state, disallowed;
+
+  if (self->resize_tick_id != 0)
+    return;
+
+  toplevel_state = gdk_toplevel_get_state (GDK_TOPLEVEL (
+      gtk_native_get_surface (GTK_NATIVE (self))));
+  disallowed = (GDK_TOPLEVEL_STATE_MINIMIZED
+      | GDK_TOPLEVEL_STATE_MAXIMIZED
+      | GDK_TOPLEVEL_STATE_FULLSCREEN
+      | GDK_TOPLEVEL_STATE_TILED);
+
+  if ((toplevel_state & disallowed) > 0) {
+    GST_DEBUG_OBJECT (self, "Cannot resize window in disallowed state");
+    return;
+  }
+
+  player = clapper_app_window_get_player (self);
+  vstreams = clapper_player_get_video_streams (player);
+  vstream = CLAPPER_VIDEO_STREAM_CAST (
+      clapper_stream_list_get_current_stream (vstreams));
+
+  if (vstream) {
+    gint video_width = clapper_video_stream_get_width (vstream);
+    gint video_height = clapper_video_stream_get_height (vstream);
+
+    if (G_LIKELY (video_width > 0 && video_height > 0)) {
+      gint win_width, win_height, dest_width, dest_height;
+
+      gtk_window_get_default_size (GTK_WINDOW (self), &win_width, &win_height);
+
+      _calculate_win_resize (win_width, win_height,
+          video_width, video_height, &dest_width, &dest_height);
+
+      /* Only begin resize when not already at perfect size */
+      if (dest_width != win_width || dest_height != win_height) {
+        ClapperAppWindowResizeData *resize_data;
+
+        resize_data = g_new0 (ClapperAppWindowResizeData, 1);
+        resize_data->dest_width = dest_width;
+        resize_data->dest_height = dest_height;
+
+        GST_DEBUG_OBJECT (self, "Window resize start, dest: %ix%i",
+            resize_data->dest_width, resize_data->dest_height);
+
+        self->resize_tick_id = gtk_widget_add_tick_callback (GTK_WIDGET (self),
+            (GtkTickCallback) _resize_tick, resize_data, g_free);
+      }
+    }
+
+    gst_object_unref (vstream);
+  }
+}
+
+static void
+_handle_middle_click (ClapperAppWindow *self, GtkGestureClick *click)
+{
+  _resize_window (self);
+  gtk_gesture_set_state (GTK_GESTURE (click), GTK_EVENT_SEQUENCE_CLAIMED);
+}
+
+static void
+_handle_right_click (ClapperAppWindow *self, GtkGestureClick *click)
 {
   GdkSurface *surface;
   GdkEventSequence *sequence;
@@ -267,6 +459,22 @@ right_click_released_cb (GtkGestureClick *click, gint n_press,
   gtk_gesture_set_state (GTK_GESTURE (click), GTK_EVENT_SEQUENCE_CLAIMED);
 }
 
+static void
+click_released_cb (GtkGestureClick *click, gint n_press,
+    gdouble x, gdouble y, ClapperAppWindow *self)
+{
+  switch (gtk_gesture_single_get_current_button (GTK_GESTURE_SINGLE (click))) {
+    case GDK_BUTTON_MIDDLE:
+      _handle_middle_click (self, click);
+      break;
+    case GDK_BUTTON_SECONDARY:
+      _handle_right_click (self, click);
+      break;
+    default:
+      break;
+  }
+}
+
 static void
 drag_begin_cb (GtkGestureDrag *drag,
     gdouble start_x, gdouble start_y, ClapperAppWindow *self)
@@ -866,6 +1074,23 @@ toggle_fullscreen (GSimpleAction *action, GVariant *param, gpointer user_data)
   video_toggle_fullscreen_cb (CLAPPER_GTK_VIDEO_CAST (self->video), self);
 }
 
+static void
+unfullscreen (GSimpleAction *action, GVariant *param, gpointer user_data)
+{
+  GtkWindow *window = GTK_WINDOW (user_data);
+
+  if (gtk_window_is_fullscreen (window)) {
+    ClapperAppWindow *self = CLAPPER_APP_WINDOW_CAST (window);
+    video_toggle_fullscreen_cb (CLAPPER_GTK_VIDEO_CAST (self->video), self);
+  }
+}
+
+static void
+auto_resize (GSimpleAction *action, GVariant *param, gpointer user_data)
+{
+  _resize_window (CLAPPER_APP_WINDOW_CAST (user_data));
+}
+
 static void
 show_help_overlay (GSimpleAction *action, GVariant *param, gpointer user_data)
 {
@@ -903,6 +1128,13 @@ clapper_app_window_get_player (ClapperAppWindow *self)
   return clapper_gtk_video_get_player (CLAPPER_GTK_VIDEO_CAST (self->video));
 }
 
+ClapperAppWindowExtraOptions *
+clapper_app_window_get_extra_options (ClapperAppWindow *self)
+{
+  return g_object_get_qdata ((GObject *) self,
+      clapper_app_window_extra_options_get_quark ());
+}
+
 void
 clapper_app_window_ensure_no_initial_state (ClapperAppWindow *self)
 {
@@ -964,10 +1196,21 @@ clapper_app_window_unrealize (GtkWidget *widget)
 static void
 clapper_app_window_init (ClapperAppWindow *self)
 {
+  ClapperAppWindowExtraOptions *extra_opts;
   GtkSettings *settings;
   GtkWidget *dummy_titlebar;
   gint distance = 0;
 
+  gtk_widget_set_size_request (GTK_WIDGET (self),
+      MIN_WINDOW_WIDTH, MIN_WINDOW_HEIGHT);
+
+  extra_opts = g_new0 (ClapperAppWindowExtraOptions, 1);
+  GST_TRACE ("Created window extra options: %p", extra_opts);
+
+  g_object_set_qdata_full ((GObject *) self,
+      clapper_app_window_extra_options_get_quark (),
+      extra_opts, (GDestroyNotify) clapper_app_window_extra_options_free);
+
   gtk_widget_init_template (GTK_WIDGET (self));
 
   /* Make double tap easier to perform */
@@ -1001,20 +1244,26 @@ clapper_app_window_constructed (GObject *object)
 
   static const GActionEntry win_entries[] = {
     { "toggle-fullscreen", toggle_fullscreen, NULL, NULL, NULL },
+    { "unfullscreen", unfullscreen, NULL, NULL, NULL },
+    { "auto-resize", auto_resize, NULL, NULL, NULL },
     { "show-help-overlay", show_help_overlay, NULL, NULL, NULL },
   };
 
 #if (CLAPPER_HAVE_MPRIS || CLAPPER_HAVE_SERVER || CLAPPER_HAVE_DISCOVERER)
   ClapperFeature *feature = NULL;
 #endif
+#if CLAPPER_HAVE_MPRIS
+  gchar mpris_name[45];
+  g_snprintf (mpris_name, sizeof (mpris_name),
+      "org.mpris.MediaPlayer2.Clapper.instance%" G_GUINT16_FORMAT, instance_count++);
+#endif
 
   self->settings = g_settings_new (CLAPPER_APP_ID);
   self->last_volume = PERCENTAGE_ROUND (g_settings_get_double (self->settings, "volume"));
 
 #if CLAPPER_HAVE_MPRIS
   feature = CLAPPER_FEATURE (clapper_mpris_new (
-      "org.mpris.MediaPlayer2.Clapper",
-      "Clapper", CLAPPER_APP_ID));
+      mpris_name, CLAPPER_APP_NAME, CLAPPER_APP_ID));
   clapper_mpris_set_queue_controllable (CLAPPER_MPRIS (feature), TRUE);
   clapper_player_add_feature (player, feature);
   gst_object_unref (feature);
@@ -1040,10 +1289,12 @@ clapper_app_window_constructed (GObject *object)
 
   clapper_player_set_autoplay (player, TRUE);
 
-  /* No need to also call this here, as item is selected
+  /* No need to also call these here, as they only change
    * after application window is contructed */
   g_signal_connect (queue, "notify::current-item",
       G_CALLBACK (_queue_current_item_changed_cb), self);
+  g_signal_connect (player, "notify::adaptive-bandwidth",
+      G_CALLBACK (_player_adaptive_bandwidth_changed_cb), NULL);
 
   g_settings_bind (self->settings, "audio-offset",
       player, "audio-offset", G_SETTINGS_BIND_GET);
@@ -1081,6 +1332,11 @@ clapper_app_window_dispose (GObject *object)
 {
   ClapperAppWindow *self = CLAPPER_APP_WINDOW_CAST (object);
 
+  if (self->resize_tick_id != 0) {
+    gtk_widget_remove_tick_callback (GTK_WIDGET (self), self->resize_tick_id);
+    self->resize_tick_id = 0;
+  }
+
   g_clear_handle_id (&self->seek_timeout, g_source_remove);
 
   gtk_widget_dispose_template (GTK_WIDGET (object), CLAPPER_APP_TYPE_WINDOW);
@@ -1140,8 +1396,8 @@ clapper_app_window_class_init (ClapperAppWindowClass *klass)
   gtk_widget_class_bind_template_callback (widget_class, key_pressed_cb);
   gtk_widget_class_bind_template_callback (widget_class, key_released_cb);
 
-  gtk_widget_class_bind_template_callback (widget_class, right_click_pressed_cb);
-  gtk_widget_class_bind_template_callback (widget_class, right_click_released_cb);
+  gtk_widget_class_bind_template_callback (widget_class, click_pressed_cb);
+  gtk_widget_class_bind_template_callback (widget_class, click_released_cb);
   gtk_widget_class_bind_template_callback (widget_class, drag_begin_cb);
   gtk_widget_class_bind_template_callback (widget_class, drag_update_cb);
 
