@@ -101,6 +101,7 @@
 #include "clapper-gtk-lead-container.h"
 #include "clapper-gtk-status-private.h"
 #include "clapper-gtk-buffering-animation-private.h"
+#include "clapper-gtk-video-placeholder-private.h"
 
 #define PERCENTAGE_ROUND(a) (round ((gdouble) a / 0.01) * 0.01)
 
@@ -1032,10 +1033,8 @@ _video_sink_changed_cb (ClapperPlayer *player,
   }
 
   if (!widget) {
-    /* FIXME: Create some default widget to show */
-    widget = g_object_ref_sink (gtk_box_new (GTK_ORIENTATION_VERTICAL, 0));
-
     GST_DEBUG_OBJECT (self, "No widget from video sink, using placeholder");
+    widget = g_object_ref_sink (clapper_gtk_video_placeholder_new ());
   }
 
   gtk_overlay_set_child (GTK_OVERLAY (self->overlay), widget);
@@ -1065,7 +1064,7 @@ _player_missing_plugin_cb (ClapperPlayer *player, const gchar *name,
 {
   /* Some media files have custom/proprietary metadata,
    * it should be safe to simply ignore these */
-  if (g_str_has_prefix (name, "meta/"))
+  if (strstr (name, "meta/") != NULL)
     return;
 
   /* XXX: Playbin2 seems to not emit state change here,
@@ -1583,7 +1582,7 @@ clapper_gtk_video_class_init (ClapperGtkVideoClass *klass)
    */
   param_specs[PROP_FADE_DELAY] = g_param_spec_uint ("fade-delay",
       NULL, NULL, 1, G_MAXUINT, DEFAULT_FADE_DELAY,
-      G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
    * ClapperGtkVideo:touch-fade-delay:
@@ -1593,7 +1592,7 @@ clapper_gtk_video_class_init (ClapperGtkVideoClass *klass)
    */
   param_specs[PROP_TOUCH_FADE_DELAY] = g_param_spec_uint ("touch-fade-delay",
       NULL, NULL, 1, G_MAXUINT, DEFAULT_TOUCH_FADE_DELAY,
-      G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
   /**
    * ClapperGtkVideo:auto-inhibit:
