@@ -50,6 +50,8 @@ struct _ClapperMediaItem
   gchar *container_format;
   gdouble duration;
 
+  gchar *cache_uri;
+
   /* For shuffle */
   gboolean used;
 };
@@ -60,6 +62,7 @@ enum
   PROP_ID,
   PROP_URI,
   PROP_SUBURI,
+  PROP_CACHE_LOCATION,
   PROP_TITLE,
   PROP_CONTAINER_FORMAT,
   PROP_DURATION,
@@ -141,6 +144,35 @@ clapper_media_item_new_from_file (GFile *file)
   return item;
 }
 
+/**
+ * clapper_media_item_new_cached:
+ * @uri: a media URI
+ * @location: (type filename) (nullable): a path to downloaded file
+ *
+ * Same as [ctor@Clapper.MediaItem.new], but allows to provide
+ * a location of a cache file where particular media at @uri
+ * is supposed to be found.
+ *
+ * File at @location existence will be checked upon starting playback
+ * of created item. If cache file is not found, media item @uri will be
+ * used as fallback. In this case when [property@Clapper.Player:download-enabled]
+ * is set to %TRUE, item will be downloaded and cached again if possible.
+ *
+ * Returns: (transfer full): a new #ClapperMediaItem.
+ *
+ * Since: 0.8
+ */
+ClapperMediaItem *
+clapper_media_item_new_cached (const gchar *uri, const gchar *location)
+{
+  ClapperMediaItem *item = clapper_media_item_new (uri);
+
+  if (location && G_LIKELY (item != NULL))
+    clapper_media_item_set_cache_location (item, location);
+
+  return item;
+}
+
 /**
  * clapper_media_item_get_id:
  * @item: a #ClapperMediaItem
@@ -450,6 +482,43 @@ clapper_media_item_update_from_discoverer_info (ClapperMediaItem *self, GstDisco
   gst_object_unref (player);
 }
 
+/* XXX: Must be set from player thread or upon construction */
+void
+clapper_media_item_set_cache_location (ClapperMediaItem *self, const gchar *location)
+{
+  g_clear_pointer (&self->cache_uri, g_free);
+
+  if (location)
+    self->cache_uri = g_filename_to_uri (location, NULL, NULL);
+
+  GST_DEBUG_OBJECT (self, "Set cache URI: \"%s\"",
+      GST_STR_NULL (self->cache_uri));
+}
+
+/* XXX: Can only be read from player thread.
+ * Returns cache URI if available, item URI otherwise. */
+inline const gchar *
+clapper_media_item_get_playback_uri (ClapperMediaItem *self)
+{
+  if (self->cache_uri) {
+    GFile *file = g_file_new_for_uri (self->cache_uri);
+    gboolean exists;
+
+    /* It is an app error if it removes files in non-stopped state,
+     * and this function is only called when starting playback */
+    exists = g_file_query_exists (file, NULL);
+    g_object_unref (file);
+
+    if (exists)
+      return self->cache_uri;
+
+    /* Do not test file existence next time */
+    clapper_media_item_set_cache_location (self, NULL);
+  }
+
+  return self->uri;
+}
+
 void
 clapper_media_item_set_used (ClapperMediaItem *self, gboolean used)
 {
@@ -505,6 +574,8 @@ clapper_media_item_finalize (GObject *object)
   gst_object_unparent (GST_OBJECT_CAST (self->timeline));
   gst_object_unref (self->timeline);
 
+  g_free (self->cache_uri);
+
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -521,6 +592,9 @@ clapper_media_item_set_property (GObject *object, guint prop_id,
     case PROP_SUBURI:
       clapper_media_item_set_suburi (self, g_value_get_string (value));
       break;
+    case PROP_CACHE_LOCATION:
+      clapper_media_item_set_cache_location (self, g_value_get_string (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -601,6 +675,17 @@ clapper_media_item_class_init (ClapperMediaItemClass *klass)
       NULL, NULL, NULL,
       G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * ClapperMediaItem:cache-location:
+   *
+   * Media downloaded cache file location.
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_CACHE_LOCATION] = g_param_spec_string ("cache-location",
+      NULL, NULL, NULL,
+      G_PARAM_WRITABLE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
   /**
    * ClapperMediaItem:title:
    *
