@@ -30,6 +30,12 @@
 #include "clapper-stream-private.h"
 #include "clapper-stream-list-private.h"
 
+#include "clapper-functionalities-availability.h"
+
+#if CLAPPER_WITH_ENHANCERS_LOADER
+#include "gst/clapper-enhancer-src-private.h"
+#endif
+
 #define GST_CAT_DEFAULT clapper_playbin_bus_debug
 GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);
 
@@ -749,7 +755,7 @@ _handle_stream_change_msg (GstMessage *msg,
           && player->current_state >= GST_STATE_PAUSED) {
         /* XXX: I am not sure if we "officially" need to flush seek after select
          * streams, but as of GStreamer 1.22 it doesn't work otherwise. */
-        _perform_flush_seek (player);
+        player->pending_flush = TRUE;
       }
       g_list_free (list);
     }
@@ -827,6 +833,40 @@ _handle_element_msg (GstMessage *msg, ClapperPlayer *player)
 
     g_free (name);
     g_free (details);
+  } else if (gst_message_has_name (msg, "GstCacheDownloadComplete")) {
+    ClapperMediaItem *downloaded_item = NULL;
+    const GstStructure *structure;
+    const gchar *location;
+    guint signal_id;
+
+    GST_OBJECT_LOCK (player);
+
+    /* Short video might be fully downloaded before playback starts */
+    if (player->pending_item)
+      downloaded_item = gst_object_ref (player->pending_item);
+    else if (player->played_item)
+      downloaded_item = gst_object_ref (player->played_item);
+
+    GST_OBJECT_UNLOCK (player);
+
+    if (G_UNLIKELY (downloaded_item == NULL)) {
+      GST_WARNING_OBJECT (player, "Download completed without media item set");
+      return;
+    }
+
+    structure = gst_message_get_structure (msg);
+    location = gst_structure_get_string (structure, "location");
+    signal_id = g_signal_lookup ("download-complete", CLAPPER_TYPE_PLAYER);
+
+    GST_INFO_OBJECT (player, "Download of %" GST_PTR_FORMAT
+        " complete: %s", downloaded_item, location);
+    clapper_media_item_set_cache_location (downloaded_item, location);
+
+    clapper_app_bus_post_object_desc_signal (player->app_bus,
+        GST_OBJECT_CAST (player), signal_id,
+        GST_OBJECT_CAST (downloaded_item), location);
+
+    gst_object_unref (downloaded_item);
   }
 }
 
@@ -835,6 +875,7 @@ _handle_tag_msg (GstMessage *msg, ClapperPlayer *player)
 {
   GstObject *src = GST_MESSAGE_SRC (msg);
   GstTagList *tags = NULL;
+  gboolean from_enhancer_src;
 
   /* Tag messages should only be posted by sink elements */
   if (G_UNLIKELY (!src))
@@ -845,8 +886,21 @@ _handle_tag_msg (GstMessage *msg, ClapperPlayer *player)
   GST_LOG_OBJECT (player, "Got tags from element: %s: %" GST_PTR_FORMAT,
       GST_OBJECT_NAME (src), tags);
 
-  if (G_LIKELY (player->played_item != NULL))
+#if CLAPPER_WITH_ENHANCERS_LOADER
+  from_enhancer_src = CLAPPER_IS_ENHANCER_SRC (src);
+#else
+  from_enhancer_src = FALSE;
+#endif
+
+  /* ClapperEnhancerSrc determines tags before stream start */
+  if (from_enhancer_src) {
+    if (player->pending_tags) {
+      gst_tag_list_unref (player->pending_tags);
+    }
+    player->pending_tags = gst_tag_list_ref (tags);
+  } else if (G_LIKELY (player->played_item != NULL)) {
     clapper_media_item_update_from_tag_list (player->played_item, tags, player);
+  }
 
   gst_tag_list_unref (tags);
 }
@@ -856,11 +910,10 @@ _handle_toc_msg (GstMessage *msg, ClapperPlayer *player)
 {
   GstObject *src = GST_MESSAGE_SRC (msg);
   GstToc *toc = NULL;
-  ClapperTimeline *timeline;
-  gboolean updated = FALSE;
+  gboolean from_enhancer_src, updated = FALSE;
 
-  /* TOC messages should only be posted by sink elements after start */
-  if (G_UNLIKELY (!src || !player->played_item))
+  /* TOC messages should only be posted by sink elements */
+  if (G_UNLIKELY (!src))
     return;
 
   /* Either new TOC was found or previous one was updated */
@@ -870,11 +923,27 @@ _handle_toc_msg (GstMessage *msg, ClapperPlayer *player)
       " from element: %s, updated: %s",
       toc, GST_OBJECT_NAME (src), (updated) ? "yes" : "no");
 
-  timeline = clapper_media_item_get_timeline (player->played_item);
+#if CLAPPER_WITH_ENHANCERS_LOADER
+  from_enhancer_src = CLAPPER_IS_ENHANCER_SRC (src);
+#else
+  from_enhancer_src = FALSE;
+#endif
 
-  if (clapper_timeline_set_toc (timeline, toc, updated)) {
-    clapper_app_bus_post_refresh_timeline (player->app_bus,
-        GST_OBJECT_CAST (player->played_item));
+  /* ClapperEnhancerSrc determines TOC before stream start */
+  if (from_enhancer_src) {
+    if (player->pending_toc) {
+      gst_toc_unref (player->pending_toc);
+    }
+    player->pending_toc = gst_toc_ref (toc);
+  } else if (G_LIKELY (player->played_item != NULL)) {
+    ClapperTimeline *timeline;
+
+    timeline = clapper_media_item_get_timeline (player->played_item);
+
+    if (clapper_timeline_set_toc (timeline, toc, updated)) {
+      clapper_app_bus_post_refresh_timeline (player->app_bus,
+          GST_OBJECT_CAST (player->played_item));
+    }
   }
 
   gst_toc_unref (toc);
@@ -952,6 +1021,13 @@ _handle_streams_selected_msg (GstMessage *msg, ClapperPlayer *player)
      * we iterate in search for all active ones */
     clapper_player_playbin_update_current_decoders (player);
   }
+
+  if (player->pending_flush) {
+    player->pending_flush = FALSE;
+
+    if (player->current_state >= GST_STATE_PAUSED)
+      _perform_flush_seek (player);
+  }
 }
 
 static inline void
@@ -1001,6 +1077,26 @@ _handle_stream_start_msg (GstMessage *msg, ClapperPlayer *player)
   /* With playbin2 we update all decoders at once after stream start */
   if (!player->use_playbin3)
     clapper_player_playbin_update_current_decoders (player);
+
+  if (player->pending_tags) {
+    if (G_LIKELY (player->played_item != NULL))
+      clapper_media_item_update_from_tag_list (player->played_item, player->pending_tags, player);
+
+    gst_clear_tag_list (&player->pending_tags);
+  }
+  if (player->pending_toc) {
+    if (G_LIKELY (player->played_item != NULL)) {
+      ClapperTimeline *timeline = clapper_media_item_get_timeline (player->played_item);
+
+      if (clapper_timeline_set_toc (timeline, player->pending_toc, FALSE)) {
+        clapper_app_bus_post_refresh_timeline (player->app_bus,
+            GST_OBJECT_CAST (player->played_item));
+      }
+    }
+
+    gst_toc_unref (player->pending_toc);
+    player->pending_toc = NULL;
+  }
 }
 
 static inline void
