@@ -43,7 +43,7 @@
 #include "clapper-playbin-bus-private.h"
 #include "clapper-app-bus-private.h"
 #include "clapper-queue-private.h"
-#include "clapper-media-item.h"
+#include "clapper-media-item-private.h"
 #include "clapper-stream-list-private.h"
 #include "clapper-stream-private.h"
 #include "clapper-video-stream-private.h"
@@ -61,6 +61,8 @@
 #define DEFAULT_VIDEO_ENABLED TRUE
 #define DEFAULT_AUDIO_ENABLED TRUE
 #define DEFAULT_SUBTITLES_ENABLED TRUE
+#define DEFAULT_DOWNLOAD_ENABLED FALSE
+#define DEFAULT_ADAPTIVE_START_BITRATE 1600000
 
 #define GST_CAT_DEFAULT clapper_player_debug
 GST_DEBUG_CATEGORY_STATIC (GST_CAT_DEFAULT);
@@ -90,6 +92,12 @@ enum
   PROP_VIDEO_ENABLED,
   PROP_AUDIO_ENABLED,
   PROP_SUBTITLES_ENABLED,
+  PROP_DOWNLOAD_DIR,
+  PROP_DOWNLOAD_ENABLED,
+  PROP_ADAPTIVE_START_BITRATE,
+  PROP_ADAPTIVE_MIN_BITRATE,
+  PROP_ADAPTIVE_MAX_BITRATE,
+  PROP_ADAPTIVE_BANDWIDTH,
   PROP_AUDIO_OFFSET,
   PROP_SUBTITLE_OFFSET,
   PROP_SUBTITLE_FONT_DESC,
@@ -99,6 +107,7 @@ enum
 enum
 {
   SIGNAL_SEEK_DONE,
+  SIGNAL_DOWNLOAD_COMPLETE,
   SIGNAL_MISSING_PLUGIN,
   SIGNAL_WARNING,
   SIGNAL_ERROR,
@@ -278,14 +287,15 @@ void
 clapper_player_handle_playbin_flags_changed (ClapperPlayer *self, const GValue *value)
 {
   gint flags;
-  gboolean video_enabled, audio_enabled, subtitles_enabled;
-  gboolean video_changed, audio_changed, subtitles_changed;
+  gboolean video_enabled, audio_enabled, subtitles_enabled, download_enabled;
+  gboolean video_changed, audio_changed, subtitles_changed, download_changed;
 
   flags = g_value_get_flags (value);
 
   video_enabled = ((flags & CLAPPER_PLAYER_PLAY_FLAG_VIDEO) == CLAPPER_PLAYER_PLAY_FLAG_VIDEO);
   audio_enabled = ((flags & CLAPPER_PLAYER_PLAY_FLAG_AUDIO) == CLAPPER_PLAYER_PLAY_FLAG_AUDIO);
   subtitles_enabled = ((flags & CLAPPER_PLAYER_PLAY_FLAG_TEXT) == CLAPPER_PLAYER_PLAY_FLAG_TEXT);
+  download_enabled = ((flags & CLAPPER_PLAYER_PLAY_FLAG_DOWNLOAD) == CLAPPER_PLAYER_PLAY_FLAG_DOWNLOAD);
 
   GST_OBJECT_LOCK (self);
 
@@ -295,6 +305,8 @@ clapper_player_handle_playbin_flags_changed (ClapperPlayer *self, const GValue *
     self->audio_enabled = audio_enabled;
   if ((subtitles_changed = self->subtitles_enabled != subtitles_enabled))
     self->subtitles_enabled = subtitles_enabled;
+  if ((download_changed = self->download_enabled != download_enabled))
+    self->download_enabled = download_enabled;
 
   GST_OBJECT_UNLOCK (self);
 
@@ -313,6 +325,11 @@ clapper_player_handle_playbin_flags_changed (ClapperPlayer *self, const GValue *
     clapper_app_bus_post_prop_notify (self->app_bus,
         GST_OBJECT_CAST (self), param_specs[PROP_SUBTITLES_ENABLED]);
   }
+  if (download_changed) {
+    GST_INFO_OBJECT (self, "Download enabled: %s", (download_enabled) ? "yes" : "no");
+    clapper_app_bus_post_prop_notify (self->app_bus,
+        GST_OBJECT_CAST (self), param_specs[PROP_DOWNLOAD_ENABLED]);
+  }
 }
 
 void
@@ -438,7 +455,7 @@ clapper_player_set_pending_item (ClapperPlayer *self, ClapperMediaItem *pending_
 
   /* Might be NULL (e.g. after queue is cleared) */
   if (pending_item) {
-    uri = clapper_media_item_get_uri (pending_item);
+    uri = clapper_media_item_get_playback_uri (pending_item);
     suburi = clapper_media_item_get_suburi (pending_item);
   }
 
@@ -700,6 +717,32 @@ clapper_player_playbin_update_current_decoders (ClapperPlayer *self)
     GST_DEBUG_OBJECT (self, "Active audio decoder not found");
 }
 
+static void
+_adaptive_demuxer_bandwidth_changed_cb (GstElement *adaptive_demuxer,
+    GParamSpec *pspec G_GNUC_UNUSED, ClapperPlayer *self)
+{
+  guint bandwidth = 0;
+  gboolean changed;
+
+  g_object_get (adaptive_demuxer, "current-bandwidth", &bandwidth, NULL);
+
+  /* Skip uncalculated bandwidth from
+   * new adaptive demuxer instance */
+  if (bandwidth == 0)
+    return;
+
+  GST_OBJECT_LOCK (self);
+  if ((changed = bandwidth != self->bandwidth))
+    self->bandwidth = bandwidth;
+  GST_OBJECT_UNLOCK (self);
+
+  if (changed) {
+    GST_LOG_OBJECT (self, "Adaptive bandwidth: %u", bandwidth);
+    clapper_app_bus_post_prop_notify (self->app_bus,
+        GST_OBJECT_CAST (self), param_specs[PROP_ADAPTIVE_BANDWIDTH]);
+  }
+}
+
 void
 clapper_player_reset (ClapperPlayer *self, gboolean pending_dispose)
 {
@@ -708,6 +751,7 @@ clapper_player_reset (ClapperPlayer *self, gboolean pending_dispose)
   GST_DEBUG_OBJECT (self, "Reset");
 
   self->had_error = FALSE;
+  self->pending_flush = FALSE;
   gst_clear_object (&self->played_item);
 
   if (pending_dispose) {
@@ -715,8 +759,21 @@ clapper_player_reset (ClapperPlayer *self, gboolean pending_dispose)
     gst_clear_object (&self->audio_decoder);
   }
 
+  if (self->adaptive_demuxer) {
+    g_signal_handlers_disconnect_by_func (self->adaptive_demuxer,
+        _adaptive_demuxer_bandwidth_changed_cb, self);
+    gst_clear_object (&self->adaptive_demuxer);
+  }
+
   GST_OBJECT_UNLOCK (self);
 
+  gst_clear_tag_list (&self->pending_tags);
+
+  if (self->pending_toc) {
+    gst_toc_unref (self->pending_toc);
+    self->pending_toc = NULL;
+  }
+
   /* Emit notify when we are not going to be disposed */
   if (!pending_dispose) {
     /* Clear current decoders (next item might not have video/audio track) */
@@ -725,6 +782,83 @@ clapper_player_reset (ClapperPlayer *self, gboolean pending_dispose)
   }
 }
 
+static inline gchar *
+_make_download_template (ClapperPlayer *self)
+{
+  gchar *download_template = NULL;
+
+  GST_OBJECT_LOCK (self);
+
+  if (self->download_enabled && self->download_dir) {
+    if (g_mkdir_with_parents (self->download_dir, 0755) == 0) {
+      download_template = g_build_filename (self->download_dir, "XXXXXX", NULL);
+    } else {
+      GST_ERROR_OBJECT (self, "Could not create download dir: \"%s\"", self->download_dir);
+    }
+  }
+
+  GST_OBJECT_UNLOCK (self);
+
+  return download_template;
+}
+
+static void
+_element_setup_cb (GstElement *playbin, GstElement *element, ClapperPlayer *self)
+{
+  GstElementFactory *factory = gst_element_get_factory (element);
+  const gchar *factory_name;
+
+  if (G_UNLIKELY (factory == NULL))
+    return;
+
+  factory_name = g_intern_static_string (GST_OBJECT_NAME (factory));
+  GST_INFO_OBJECT (self, "Element setup: %s", factory_name);
+
+  if (factory_name == g_intern_static_string ("downloadbuffer")) {
+    gchar *download_template;
+
+    /* Only set props if we have download template */
+    if ((download_template = _make_download_template (self))) {
+      g_object_set (element,
+          "temp-template", download_template,
+          "temp-remove", FALSE,
+          NULL);
+      g_free (download_template);
+    }
+  } else if (factory_name == g_intern_static_string ("dashdemux2")
+      || factory_name == g_intern_static_string ("hlsdemux2")) {
+    guint start_bitrate, min_bitrate, max_bitrate;
+
+    GST_OBJECT_LOCK (self);
+
+    start_bitrate = self->start_bitrate;
+    min_bitrate = self->min_bitrate;
+    max_bitrate = self->max_bitrate;
+
+    if (self->adaptive_demuxer) {
+      g_signal_handlers_disconnect_by_func (self->adaptive_demuxer,
+          _adaptive_demuxer_bandwidth_changed_cb, self);
+    }
+
+    gst_object_replace ((GstObject **) &self->adaptive_demuxer, GST_OBJECT_CAST (element));
+
+    if (self->adaptive_demuxer) {
+      g_signal_connect (self->adaptive_demuxer, "notify::current-bandwidth",
+          G_CALLBACK (_adaptive_demuxer_bandwidth_changed_cb), self);
+    }
+
+    GST_OBJECT_UNLOCK (self);
+
+    g_object_set (element,
+        "low-watermark-time", 3 * GST_SECOND,
+        "high-watermark-time", 10 * GST_SECOND,
+        "start-bitrate", start_bitrate,
+        "min-bitrate", min_bitrate,
+        "max-bitrate", max_bitrate,
+        NULL);
+  }
+}
+
 static void
 _about_to_finish_cb (GstElement *playbin, ClapperPlayer *self)
 {
@@ -1505,6 +1639,293 @@ clapper_player_get_subtitles_enabled (ClapperPlayer *self)
   return enabled;
 }
 
+/**
+ * clapper_player_set_download_dir:
+ * @player: a #ClapperPlayer
+ * @path: (type filename): the path of a directory to use for media downloads
+ *
+ * Set a directory that @player will use to store downloads.
+ *
+ * See [property@Clapper.Player:download-enabled] description for more
+ * info how this works.
+ *
+ * Since: 0.8
+ */
+void
+clapper_player_set_download_dir (ClapperPlayer *self, const gchar *path)
+{
+  gboolean changed;
+
+  g_return_if_fail (CLAPPER_IS_PLAYER (self));
+  g_return_if_fail (path != NULL);
+
+  GST_OBJECT_LOCK (self);
+  changed = g_set_str (&self->download_dir, path);
+  GST_OBJECT_UNLOCK (self);
+
+  if (changed) {
+    GST_INFO_OBJECT (self, "Current download dir: %s", path);
+    clapper_app_bus_post_prop_notify (self->app_bus,
+        GST_OBJECT_CAST (self), param_specs[PROP_DOWNLOAD_DIR]);
+  }
+}
+
+/**
+ * clapper_player_get_download_dir:
+ * @player: a #ClapperPlayer
+ *
+ * Get path to a directory set for media downloads.
+ *
+ * Returns: (type filename) (transfer full) (nullable): the path of a directory
+ *   set for media downloads or %NULL if no directory was set yet.
+ *
+ * Since: 0.8
+ */
+gchar *
+clapper_player_get_download_dir (ClapperPlayer *self)
+{
+  gchar *download_dir;
+
+  g_return_val_if_fail (CLAPPER_IS_PLAYER (self), NULL);
+
+  GST_OBJECT_LOCK (self);
+  download_dir = g_strdup (self->download_dir);
+  GST_OBJECT_UNLOCK (self);
+
+  return download_dir;
+}
+
+/**
+ * clapper_player_set_download_enabled:
+ * @player: a #ClapperPlayer
+ * @enabled: whether enabled
+ *
+ * Set whether player should attempt progressive download buffering.
+ *
+ * For this to actually work a [property@Clapper.Player:download-dir]
+ * must also be set.
+ *
+ * Since: 0.8
+ */
+void
+clapper_player_set_download_enabled (ClapperPlayer *self, gboolean enabled)
+{
+  g_return_if_fail (CLAPPER_IS_PLAYER (self));
+
+  clapper_playbin_bus_post_set_play_flag (self->bus, CLAPPER_PLAYER_PLAY_FLAG_DOWNLOAD, enabled);
+}
+
+/**
+ * clapper_player_get_download_enabled:
+ * @player: a #ClapperPlayer
+ *
+ * Get whether progressive download buffering is enabled.
+ *
+ * Returns: %TRUE if enabled, %FALSE otherwise.
+ *
+ * Since: 0.8
+ */
+gboolean
+clapper_player_get_download_enabled (ClapperPlayer *self)
+{
+  gboolean enabled;
+
+  g_return_val_if_fail (CLAPPER_IS_PLAYER (self), FALSE);
+
+  GST_OBJECT_LOCK (self);
+  enabled = self->download_enabled;
+  GST_OBJECT_UNLOCK (self);
+
+  return enabled;
+}
+
+static void
+_set_adaptive_bitrate (ClapperPlayer *self, guint *internal_ptr,
+    const gchar *prop_name, guint bitrate, GParamSpec *pspec)
+{
+  GstElement *element = NULL;
+  gboolean changed;
+
+  if (!self->use_playbin3) {
+    GST_WARNING_OBJECT (self, "Setting adaptive-%s when using playbin2"
+        " has no effect", prop_name);
+  }
+
+  GST_OBJECT_LOCK (self);
+  if ((changed = (*internal_ptr != bitrate))) {
+    *internal_ptr = bitrate;
+
+    if (self->adaptive_demuxer)
+      element = gst_object_ref (self->adaptive_demuxer);
+  }
+  GST_OBJECT_UNLOCK (self);
+
+  if (changed) {
+    GST_INFO_OBJECT (self, "Set adaptive-%s: %u", prop_name, bitrate);
+
+    if (element)
+      g_object_set (element, prop_name, bitrate, NULL);
+
+    clapper_app_bus_post_prop_notify (self->app_bus, GST_OBJECT_CAST (self), pspec);
+  }
+
+  gst_clear_object (&element);
+}
+
+/**
+ * clapper_player_set_adaptive_start_bitrate:
+ * @player: a #ClapperPlayer
+ * @bitrate: a bitrate to set (bits/s)
+ *
+ * Set initial bitrate to select when starting adaptive
+ * streaming such as DASH or HLS.
+ *
+ * Since: 0.8
+ */
+void
+clapper_player_set_adaptive_start_bitrate (ClapperPlayer *self, guint bitrate)
+{
+  g_return_if_fail (CLAPPER_IS_PLAYER (self));
+
+  _set_adaptive_bitrate (self, &self->start_bitrate,
+      "start-bitrate", bitrate, param_specs[PROP_ADAPTIVE_START_BITRATE]);
+}
+
+/**
+ * clapper_player_get_adaptive_start_bitrate:
+ * @player: a #ClapperPlayer
+ *
+ * Get currently set initial bitrate (bits/s) for adaptive streaming.
+ *
+ * Returns: the start bitrate value.
+ *
+ * Since: 0.8
+ */
+guint
+clapper_player_get_adaptive_start_bitrate (ClapperPlayer *self)
+{
+  guint bitrate;
+
+  g_return_val_if_fail (CLAPPER_IS_PLAYER (self), 0);
+
+  GST_OBJECT_LOCK (self);
+  bitrate = self->start_bitrate;
+  GST_OBJECT_UNLOCK (self);
+
+  return bitrate;
+}
+
+/**
+ * clapper_player_set_adaptive_min_bitrate:
+ * @player: a #ClapperPlayer
+ * @bitrate: a bitrate to set (bits/s)
+ *
+ * Set minimal bitrate to select for adaptive streaming
+ * such as DASH or HLS.
+ *
+ * Since: 0.8
+ */
+void
+clapper_player_set_adaptive_min_bitrate (ClapperPlayer *self, guint bitrate)
+{
+  g_return_if_fail (CLAPPER_IS_PLAYER (self));
+
+  _set_adaptive_bitrate (self, &self->min_bitrate,
+      "min-bitrate", bitrate, param_specs[PROP_ADAPTIVE_MIN_BITRATE]);
+}
+
+/**
+ * clapper_player_get_adaptive_min_bitrate:
+ * @player: a #ClapperPlayer
+ *
+ * Get currently set minimal bitrate (bits/s) for adaptive streaming.
+ *
+ * Returns: the minimal bitrate value.
+ *
+ * Since: 0.8
+ */
+guint
+clapper_player_get_adaptive_min_bitrate (ClapperPlayer *self)
+{
+  guint bitrate;
+
+  g_return_val_if_fail (CLAPPER_IS_PLAYER (self), 0);
+
+  GST_OBJECT_LOCK (self);
+  bitrate = self->min_bitrate;
+  GST_OBJECT_UNLOCK (self);
+
+  return bitrate;
+}
+
+/**
+ * clapper_player_set_adaptive_max_bitrate:
+ * @player: a #ClapperPlayer
+ * @bitrate: a bitrate to set (bits/s)
+ *
+ * Set maximal bitrate to select for adaptive streaming
+ * such as DASH or HLS.
+ *
+ * Since: 0.8
+ */
+void
+clapper_player_set_adaptive_max_bitrate (ClapperPlayer *self, guint bitrate)
+{
+  g_return_if_fail (CLAPPER_IS_PLAYER (self));
+
+  _set_adaptive_bitrate (self, &self->max_bitrate,
+      "max-bitrate", bitrate, param_specs[PROP_ADAPTIVE_MAX_BITRATE]);
+}
+
+/**
+ * clapper_player_get_adaptive_max_bitrate:
+ * @player: a #ClapperPlayer
+ *
+ * Get currently set maximal bitrate (bits/s) for adaptive streaming.
+ *
+ * Returns: the maximal bitrate value.
+ *
+ * Since: 0.8
+ */
+guint
+clapper_player_get_adaptive_max_bitrate (ClapperPlayer *self)
+{
+  guint bitrate;
+
+  g_return_val_if_fail (CLAPPER_IS_PLAYER (self), 0);
+
+  GST_OBJECT_LOCK (self);
+  bitrate = self->max_bitrate;
+  GST_OBJECT_UNLOCK (self);
+
+  return bitrate;
+}
+
+/**
+ * clapper_player_get_adaptive_bandwidth:
+ * @player: a #ClapperPlayer
+ *
+ * Get last fragment download bandwidth (bits/s) during
+ * adaptive streaming.
+ *
+ * Returns: the adaptive bandwidth.
+ *
+ * Since: 0.8
+ */
+guint
+clapper_player_get_adaptive_bandwidth (ClapperPlayer *self)
+{
+  guint bandwidth;
+
+  g_return_val_if_fail (CLAPPER_IS_PLAYER (self), 0);
+
+  GST_OBJECT_LOCK (self);
+  bandwidth = self->bandwidth;
+  GST_OBJECT_UNLOCK (self);
+
+  return bandwidth;
+}
+
 /**
  * clapper_player_set_audio_offset:
  * @player: a #ClapperPlayer
@@ -1779,7 +2200,7 @@ clapper_player_thread_start (ClapperThreadedObject *threaded_object)
     if (!(env = g_getenv ("CLAPPER_USE_PLAYBIN3"))) // Clapper override
       env = g_getenv ("GST_CLAPPER_USE_PLAYBIN3"); // compat
 
-  self->use_playbin3 = (env && g_str_has_prefix (env, "1"));
+  self->use_playbin3 = (!env || g_str_has_prefix (env, "1"));
   playbin_str = (self->use_playbin3) ? "playbin3" : "playbin";
 
   if (!(self->playbin = gst_element_factory_make (playbin_str, NULL))) {
@@ -1793,6 +2214,7 @@ clapper_player_thread_start (ClapperThreadedObject *threaded_object)
   for (i = 0; playbin_watchlist[i]; ++i)
     gst_element_add_property_notify_watch (self->playbin, playbin_watchlist[i], TRUE);
 
+  g_signal_connect (self->playbin, "element-setup", G_CALLBACK (_element_setup_cb), self);
   g_signal_connect (self->playbin, "about-to-finish", G_CALLBACK (_about_to_finish_cb), self);
 
   if (!self->use_playbin3) {
@@ -1866,6 +2288,8 @@ clapper_player_init (ClapperPlayer *self)
   self->video_enabled = DEFAULT_VIDEO_ENABLED;
   self->audio_enabled = DEFAULT_AUDIO_ENABLED;
   self->subtitles_enabled = DEFAULT_SUBTITLES_ENABLED;
+  self->download_enabled = DEFAULT_DOWNLOAD_ENABLED;
+  self->start_bitrate = DEFAULT_ADAPTIVE_START_BITRATE;
 }
 
 static void
@@ -1924,6 +2348,8 @@ clapper_player_finalize (GObject *object)
   gst_clear_object (&self->pending_item);
   gst_clear_object (&self->played_item);
 
+  g_free (self->download_dir);
+
   G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -1991,6 +2417,24 @@ clapper_player_get_property (GObject *object, guint prop_id,
     case PROP_SUBTITLES_ENABLED:
       g_value_set_boolean (value, clapper_player_get_subtitles_enabled (self));
       break;
+    case PROP_DOWNLOAD_DIR:
+      g_value_take_string (value, clapper_player_get_download_dir (self));
+      break;
+    case PROP_DOWNLOAD_ENABLED:
+      g_value_set_boolean (value, clapper_player_get_download_enabled (self));
+      break;
+    case PROP_ADAPTIVE_START_BITRATE:
+      g_value_set_uint (value, clapper_player_get_adaptive_start_bitrate (self));
+      break;
+    case PROP_ADAPTIVE_MIN_BITRATE:
+      g_value_set_uint (value, clapper_player_get_adaptive_min_bitrate (self));
+      break;
+    case PROP_ADAPTIVE_MAX_BITRATE:
+      g_value_set_uint (value, clapper_player_get_adaptive_max_bitrate (self));
+      break;
+    case PROP_ADAPTIVE_BANDWIDTH:
+      g_value_set_uint (value, clapper_player_get_adaptive_bandwidth (self));
+      break;
     case PROP_AUDIO_OFFSET:
       g_value_set_double (value, clapper_player_get_audio_offset (self));
       break;
@@ -2046,6 +2490,21 @@ clapper_player_set_property (GObject *object, guint prop_id,
     case PROP_SUBTITLES_ENABLED:
       clapper_player_set_subtitles_enabled (self, g_value_get_boolean (value));
       break;
+    case PROP_DOWNLOAD_DIR:
+      clapper_player_set_download_dir (self, g_value_get_string (value));
+      break;
+    case PROP_DOWNLOAD_ENABLED:
+      clapper_player_set_download_enabled (self, g_value_get_boolean (value));
+      break;
+    case PROP_ADAPTIVE_START_BITRATE:
+      clapper_player_set_adaptive_start_bitrate (self, g_value_get_uint (value));
+      break;
+    case PROP_ADAPTIVE_MIN_BITRATE:
+      clapper_player_set_adaptive_min_bitrate (self, g_value_get_uint (value));
+      break;
+    case PROP_ADAPTIVE_MAX_BITRATE:
+      clapper_player_set_adaptive_max_bitrate (self, g_value_get_uint (value));
+      break;
     case PROP_AUDIO_OFFSET:
       clapper_player_set_audio_offset (self, g_value_get_double (value));
       break;
@@ -2251,6 +2710,118 @@ clapper_player_class_init (ClapperPlayerClass *klass)
       NULL, NULL, DEFAULT_SUBTITLES_ENABLED,
       G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
 
+  /**
+   * ClapperPlayer:download-dir:
+   *
+   * A directory that @player will use to download network content
+   * when [property@Clapper.Player:download-enabled] is set to %TRUE.
+   *
+   * If directory at @path does not exist, it will be automatically created.
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_DOWNLOAD_DIR] = g_param_spec_string ("download-dir",
+      NULL, NULL, NULL,
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * ClapperPlayer:download-enabled:
+   *
+   * Whether progressive download buffering is enabled.
+   *
+   * If progressive download is enabled and [property@Clapper.Player:download-dir]
+   * is set, streamed network content will be cached to the disk space instead
+   * of memory whenever possible. This allows for faster seeking through
+   * currently played media.
+   *
+   * Not every type of content is download applicable. Mainly applies to
+   * web content that does not use adaptive streaming.
+   *
+   * Once data that media item URI points to is fully downloaded, player
+   * will emit [signal@Clapper.Player::download-complete] signal with a
+   * location of downloaded file.
+   *
+   * Playing again the exact same [class@Clapper.MediaItem] object that was
+   * previously fully downloaded will cause player to automatically use that
+   * cached file if it still exists, avoiding any further network requests.
+   *
+   * Please note that player will not delete nor manage downloaded content.
+   * It is up to application to cleanup data in created cache directory
+   * (e.g. before app exits), in order to remove any downloads that app
+   * is not going to use next time it is run and incomplete ones.
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_DOWNLOAD_ENABLED] = g_param_spec_boolean ("download-enabled",
+      NULL, NULL, DEFAULT_DOWNLOAD_ENABLED,
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * ClapperPlayer:adaptive-start-bitrate:
+   *
+   * An initial bitrate (bits/s) to select during
+   * starting adaptive streaming such as DASH or HLS.
+   *
+   * If value is higher than lowest available bitrate in streaming
+   * manifest, then lowest possible bitrate will be selected.
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_ADAPTIVE_START_BITRATE] = g_param_spec_uint ("adaptive-start-bitrate",
+      NULL, NULL, 0, G_MAXUINT, DEFAULT_ADAPTIVE_START_BITRATE,
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * ClapperPlayer:adaptive-min-bitrate:
+   *
+   * A minimal allowed bitrate (bits/s) during adaptive streaming
+   * such as DASH or HLS.
+   *
+   * Setting this will prevent streaming from entering lower qualities
+   * (even when connection speed cannot keep up). When set together with
+   * [property@Clapper.Player:adaptive-max-bitrate] it can be used to
+   * enforce some specific quality.
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_ADAPTIVE_MIN_BITRATE] = g_param_spec_uint ("adaptive-min-bitrate",
+      NULL, NULL, 0, G_MAXUINT, 0,
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * ClapperPlayer:adaptive-max-bitrate:
+   *
+   * A maximal allowed bitrate (bits/s) during adaptive streaming
+   * such as DASH or HLS (`0` for unlimited).
+   *
+   * Setting this will prevent streaming from entering qualities with
+   * higher bandwidth than the one set. When set together with
+   * [property@Clapper.Player:adaptive-min-bitrate] it can be used to
+   * enforce some specific quality.
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_ADAPTIVE_MAX_BITRATE] = g_param_spec_uint ("adaptive-max-bitrate",
+      NULL, NULL, 0, G_MAXUINT, 0,
+      G_PARAM_READWRITE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
+  /**
+   * ClapperPlayer:adaptive-bandwidth:
+   *
+   * Last fragment download bandwidth (bits/s) during adaptive streaming.
+   *
+   * This property only changes when adaptive streaming and later stays
+   * at the last value until streaming some adaptive content again.
+   *
+   * Apps can use this to determine and set an optimal value for
+   * [property@Clapper.Player:adaptive-start-bitrate].
+   *
+   * Since: 0.8
+   */
+  param_specs[PROP_ADAPTIVE_BANDWIDTH] = g_param_spec_uint ("adaptive-bandwidth",
+      NULL, NULL, 0, G_MAXUINT, 0,
+      G_PARAM_READABLE | G_PARAM_EXPLICIT_NOTIFY | G_PARAM_STATIC_STRINGS);
+
   /**
    * ClapperPlayer:audio-offset:
    *
@@ -2288,6 +2859,22 @@ clapper_player_class_init (ClapperPlayerClass *klass)
       G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
       0, NULL, NULL, NULL, G_TYPE_NONE, 0);
 
+  /**
+   * ClapperPlayer::download-complete:
+   * @player: a #ClapperPlayer
+   * @item: a #ClapperMediaItem
+   * @location: (type filename): a path to downloaded file
+   *
+   * Media was fully downloaded to local cache directory. This signal will
+   * be only emitted when progressive download buffering is enabled by
+   * setting [property@Clapper.Player:download-enabled] property to %TRUE.
+   *
+   * Since: 0.8
+   */
+  signals[SIGNAL_DOWNLOAD_COMPLETE] = g_signal_new ("download-complete",
+      G_TYPE_FROM_CLASS (klass), G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
+      0, NULL, NULL, NULL, G_TYPE_NONE, 2, CLAPPER_TYPE_MEDIA_ITEM, G_TYPE_STRING);
+
   /**
    * ClapperPlayer::missing-plugin:
    * @player: a #ClapperPlayer
